{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# Trees\n",
    "\n",
    "Tree data structure has a **root**, **branches**, and **leaves**. The difference between a tree in nature and a tree in computer science is that a tree data structure has its *root at the top* and its *leaves on the bottom*.\n",
    "\n",
    "Trees are used in many areas of computer science, including operating systems, graphics, database systems, and computer networking. Tree data structures have many things in common with their botanical cousins. \n",
    "\n",
    "```{mermaid}\n",
    "flowchart TB;\n",
    "    a((a))\n",
    "    b1((b1))\n",
    "    b2((b2))\n",
    "    c1((c1))\n",
    "    c2((c2))\n",
    "    c3((c3))\n",
    "    c4((c4))\n",
    "    c5((c5))\n",
    "    d1((d1))\n",
    "    d2((d2))\n",
    "    d3((d3))\n",
    "    d4((d4))\n",
    "    d5((d5))\n",
    "    d6((d6))\n",
    "    d7((d7))\n",
    "    e1((e1))\n",
    "    e2((e2))\n",
    "    e3((e3))\n",
    "    e4((e4))\n",
    "    e5((e5))\n",
    "    f1((f1))\n",
    "\n",
    "    a --> b1\n",
    "    a --> b2\n",
    "    b1 --> c1\n",
    "    b1 --> c2\n",
    "    b2 --> c3\n",
    "    b2 --> c4\n",
    "    b2 --> c5\n",
    "    c1 --> d1\n",
    "    c2 --> d2\n",
    "    c2 --> d3\n",
    "    c3 --> d4\n",
    "    c4 --> d5\n",
    "    c4 --> d6\n",
    "    c4 --> d7\n",
    "    d2 --> e1\n",
    "    d2 --> e2\n",
    "    d5 --> e3\n",
    "    d5 --> e4\n",
    "    d5 --> e5\n",
    "    e4 --> f1\n",
    "\n",
    "    style a fill:#FF8000\n",
    "    style c5 fill:#9eea00\n",
    "    style d1 fill:#9eea00\n",
    "    style d3 fill:#9eea00\n",
    "    style d4 fill:#9eea00\n",
    "    style d6 fill:#9eea00\n",
    "    style d7 fill:#9eea00\n",
    "    style e1 fill:#9eea00\n",
    "    style e2 fill:#9eea00\n",
    "    style e3 fill:#9eea00\n",
    "    style e5 fill:#9eea00\n",
    "    style f1 fill:#9eea00\n",
    "```\n",
    "\n",
    "**Example**\n",
    "\n",
    "1. Taxonomic classification tree from biology.\n",
    "2. Computer file system for storing data. In a file system, directories, or folders, are structured as a tree.\n",
    "3. A web page written in HTML is a tree.\n",
    "\n",
    "## Prperties of trees\n",
    "\n",
    "1. First property of trees is that trees are **hierarchical**, trees are structured in layers with the *more general things near the top and the more specific things near the bottom*.\n",
    "2. Second property of trees is that all of the *children of one node are independent of the children of another node*.\n",
    "3. Third property is that *each leaf node is unique*.\n",
    "\n",
    "If each node in the tree has a maximum of two children, we say that the tree is a **binary tree.**"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Tree Traversals\n",
    "\n",
    "Visitation of the nodes of a tree is called a **“traversal”**. The three traversals we will look at are called preorder, inorder, and postorder.\n",
    "\n",
    "\n",
    "- **Pre-Order**: visit the root node first, then recursively do preorder traversal of the left subtree, followed by recursive preorder traversal of the right subtree.\n",
    "- **In-Order**: recursively do an inorder traversal on the left subtree, visit the root node, and finally do a recursive inorder traversal of the right subtree.\n",
    "- **Post-Order**: recursively do a postorder traversal of the left subtree and the right subtree followed by a visit to the root node.\n",
    "\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 1,
   "metadata": {},
   "outputs": [],
   "source": [
    "class BinaryTree:\n",
    "    def __init__(self,obj):\n",
    "        self.key = obj\n",
    "        self.left_child = None\n",
    "        self.right_child = None\n",
    "    \n",
    "    def insert_left(self,newNode):\n",
    "        if self.left_child == None:\n",
    "            self.left_child = BinaryTree(newNode)\n",
    "        else:\n",
    "            t = BinaryTree(newNode)\n",
    "            t.left_child = self.left_child\n",
    "            self.left_child = t\n",
    "    \n",
    "    def insert_right(self,newNode):\n",
    "        if self.right_child == None:\n",
    "            self.right_child = BinaryTree(newNode)\n",
    "        else:\n",
    "            t = BinaryTree(newNode)\n",
    "            t.right_child = self.right_child\n",
    "            self.right_child = t\n",
    "    \n",
    "    def get_right_child(self):\n",
    "        return self.right_child\n",
    "\n",
    "    def get_left_child(self):\n",
    "        return self.left_child\n",
    "\n",
    "    def set_root_val(self,obj):\n",
    "        self.key = obj\n",
    "\n",
    "    def get_root_val(self):\n",
    "        return self.key\n",
    "    \n",
    "    def pre_order_traversal(self):\n",
    "        traversal = \"\"\n",
    "        traversal += f\"{self.key} \"\n",
    "\n",
    "        if self.left_child:\n",
    "            traversal += self.left_child.pre_order_traversal()\n",
    "        \n",
    "        if self.right_child:\n",
    "            traversal += self.right_child.pre_order_traversal()\n",
    "        \n",
    "        return traversal\n",
    "\n",
    "    \n",
    "    def post_order_traversal(self):\n",
    "        traversal = \"\"\n",
    "\n",
    "        if self.left_child:\n",
    "            traversal += self.left_child.pre_order_traversal()\n",
    "        \n",
    "        if self.right_child:\n",
    "            traversal += self.right_child.pre_order_traversal()\n",
    "\n",
    "        traversal += f\"{self.key} \"\n",
    "        \n",
    "        return traversal\n",
    "    \n",
    "    def in_order_traversal(self):\n",
    "        traversal = \"\"\n",
    "\n",
    "        if self.left_child:\n",
    "            traversal += self.left_child.pre_order_traversal()        \n",
    "\n",
    "        traversal += f\"{self.key} \"\n",
    "        \n",
    "        if self.right_child:\n",
    "            traversal += self.right_child.pre_order_traversal()\n",
    "        return traversal\n",
    "    "
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 2,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "a\n",
      "b\n",
      "c\n",
      "b\n",
      "d\n",
      "e\n",
      "c\n",
      "f\n",
      "g\n",
      "b d e a c f g \n",
      "a b d e c f g \n",
      "b d e c f g a \n"
     ]
    }
   ],
   "source": [
    "t = BinaryTree(\"a\")\n",
    "t.insert_left(\"b\")\n",
    "t.insert_right(\"c\")\n",
    "\n",
    "print(t.get_root_val())\n",
    "print(t.get_left_child().get_root_val())\n",
    "print(t.get_right_child().get_root_val())\n",
    "\n",
    "lc = t.get_left_child()\n",
    "lc.insert_left(\"d\")\n",
    "lc.insert_right(\"e\")\n",
    "\n",
    "print(lc.get_root_val())\n",
    "print(lc.get_left_child().get_root_val())\n",
    "print(lc.get_right_child().get_root_val())\n",
    "\n",
    "rc = t.get_right_child()\n",
    "rc.insert_left(\"f\")\n",
    "rc.insert_right(\"g\")\n",
    "\n",
    "print(rc.get_root_val())\n",
    "print(rc.get_left_child().get_root_val())\n",
    "print(rc.get_right_child().get_root_val())\n",
    "\n",
    "print(t.in_order_traversal())\n",
    "print(t.pre_order_traversal())\n",
    "print(t.post_order_traversal())"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Practice Problems"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### Write a Python function to create a parse tree for mathemetical expressions like `((3 * 2) + (12 / 4))`."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 3,
   "metadata": {},
   "outputs": [
    {
     "ename": "IndentationError",
     "evalue": "expected an indented block after function definition on line 1 (658469152.py, line 2)",
     "output_type": "error",
     "traceback": [
      "\u001b[0;36m  File \u001b[0;32m\"/var/folders/mc/bfghmnhs7v7_2zl28699c1_h0000gp/T/ipykernel_89201/658469152.py\"\u001b[0;36m, line \u001b[0;32m2\u001b[0m\n\u001b[0;31m    \u001b[0m\n\u001b[0m    ^\u001b[0m\n\u001b[0;31mIndentationError\u001b[0m\u001b[0;31m:\u001b[0m expected an indented block after function definition on line 1\n"
     ]
    }
   ],
   "source": [
    "def build_parse_tree(exp):\n",
    "    pass"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Binary Search Tree\n",
    "\n",
    "Binary search trees is another way to map from a key to a value. In this case we are not interested in the exact placement of items in the tree, but we are interested in using the binary tree structure to provide for efficient searching.\n",
    "\n",
    "Binary Search Tree is a node-based binary tree data structure which has the following properties:\n",
    "\n",
    "- The left subtree of a node contains only nodes with keys lesser than the node’s key.\n",
    "- The right subtree of a node contains only nodes with keys greater than the node’s key.\n",
    "- The left and right subtree each must also be a binary search tree."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### Binary Search Tree Abstract Data Type\n",
    "\n",
    "The Binary Search Tree abstract data type is defined by the following structure and operations.\n",
    "\n",
    "#### Binary Search Tree Operations\n",
    "\n",
    "- `BinarySearchTree()` Create a new, empty binary search tree.\n",
    "- `put(key, val)` Add a new key-value pair to the binary search tree. If the key is already in the binary search tree then replace the old value with the new value.\n",
    "- `get(key)` Given a key, return the value stored in the binary search tree or None otherwise.\n",
    "- `remove(key)` Remove the key-value pair from the binary search tree.\n",
    "- `size()` Return the number of key-value pairs stored in the binary search tree.\n",
    "\n",
    "`bst = BinarySearchTree()` creates a binary search tree that starts out empty."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "class TreeNode:\n",
    "    def __init__(self, key, value, parent):\n",
    "        self.key = key\n",
    "        self.value = value\n",
    "        self.left = None\n",
    "        self.right = None\n",
    "        self.parent = parent\n",
    "    \n",
    "    def get_key(self):\n",
    "        return self.key\n",
    "\n",
    "    def get_value(self):\n",
    "        return self.value\n",
    "\n",
    "\n",
    "class BinarySearchTree:\n",
    "    def __init__(self):\n",
    "        self.root = None\n",
    "        self.num_nodes = 0\n",
    "\n",
    "    def is_empty(self):\n",
    "        return self.root is None\n",
    "\n",
    "    def put(self, key, value):\n",
    "        self.num_nodes += 1\n",
    "        self.root = self._put(self.root, key, value)\n",
    "\n",
    "    def _put(self, node, key, value, parent = None):\n",
    "        if node is None:\n",
    "            node = TreeNode(key, value, parent)\n",
    "        else:\n",
    "            if key < node.key:\n",
    "                node.left = self._put(node.left, key, value, node)\n",
    "            elif key > node.key:\n",
    "                node.right = self._put(node.right, key, value, node)\n",
    "            else:\n",
    "                node.value = value\n",
    "\n",
    "        return node\n",
    "\n",
    "    def get(self, key):\n",
    "        return self._get(self.root, key)\n",
    "\n",
    "    def _get(self, node, key):\n",
    "        if node is None:\n",
    "            raise Exception(f\"Tree node with key {key} does not exist\")\n",
    "        else:\n",
    "            if key < node.key:\n",
    "                node = self._get(node.left, key)\n",
    "            elif key > node.key:\n",
    "                node = self._get(node.right, key)\n",
    "\n",
    "        return node\n",
    "\n",
    "    def remove(self, key):\n",
    "        self.num_nodes -= 1\n",
    "\n",
    "        node = self.get(key)\n",
    "        \n",
    "        if node.right is None and node.left is None:\n",
    "            self._reassign_nodes(node, None)\n",
    "        elif node.left is None:\n",
    "            self._reassign_nodes(node, node.right)\n",
    "        elif node.right is None:\n",
    "            self._reassign_nodes(node, node.left)\n",
    "        else:\n",
    "            lowest_node = self._get_lowest_node(node.right)\n",
    "            lowest_node.left = node.left\n",
    "            lowest_node.right = node.right\n",
    "            node.left.parent = lowest_node\n",
    "            if node.right:\n",
    "                node.right.parent = lowest_node\n",
    "            self._reassign_nodes(node, lowest_node)\n",
    "\n",
    "\n",
    "    def _reassign_nodes(self, node, new_children):\n",
    "        if new_children:\n",
    "            new_children.parent = node.parent\n",
    "\n",
    "        if node.parent:\n",
    "            if node.parent.right == node:\n",
    "                node.parent.right = new_children\n",
    "            else:\n",
    "                node.parent.left = new_children\n",
    "        else:\n",
    "            self.root = new_children\n",
    "\n",
    "    def _get_lowest_node(self, node):\n",
    "        if node.left:\n",
    "            lowest_node = self._get_lowest_node(node.left)\n",
    "        else:\n",
    "            lowest_node = node\n",
    "            self._reassign_nodes(node, node.right)\n",
    "\n",
    "        return lowest_node\n",
    "    "
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "b\n",
      "c\n",
      "h\n",
      "r\n"
     ]
    }
   ],
   "source": [
    "bst = BinarySearchTree()\n",
    "\n",
    "keys = [5, 4, 1, 2, 3, 9, 8, 7, 6, 13, 10, 14, 11]\n",
    "values = [\"a\", \"b\", \"c\", \"d\", \"e\", \"f\", \"g\", \"h\", \"i\", \"j\", \"k\", \"l\", \"m\"]\n",
    "\n",
    "for i in range(13):\n",
    "    bst.put(keys[i], values[i])\n",
    "\n",
    "print(bst.get(4).value)\n",
    "print(bst.get(1).value)\n",
    "print(bst.get(7).value)\n",
    "\n",
    "bst.put(3, \"r\")\n",
    "print(bst.get(3).value)\n",
    "# bst.get(4)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### Search Tree Analysis\n",
    "\n",
    "A perfectly balanced tree has the same number of nodes in the left subtree as the right subtree. In a balanced binary tree, the worst-case performance of put, get is `O(logn)`, where n is the number of nodes in the tree. So `log n` gives us the height of the tree, and represents the maximum number of comparisons that put will need to do as it searches for the proper place to insert a new node.\n",
    "\n",
    "Performance of the binary search tree can degrade to `O(n)` for operations like get and put when the tree becomes unbalanced.\n"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## AVL Tree\n",
    "\n",
    "AVL tree and is named for its inventors: G.M. Adelson-Velskii and E.M. Landis. **AVL Tree** is a special kind of binary search tree that automatically makes sure that the tree remains **balanced** at all times. An AVL tree implements the Map abstract data type just like a regular binary search tree, the only difference is in how the tree performs. \n",
    "\n",
    "To implement our AVL tree we need to keep track of a **balance factor** for each node in the tree, **balance factor** for each node in the tree is calculated by looking at the heights of the left and right subtrees. More formally, we define the balance factor for a node as the difference between the height of the left subtree and the height of the right subtree.\n",
    "\n",
    "    balanceFactor = height(leftSubTree) − height(rightSubTree)\n",
    "\n",
    "Using the definition for balance factor given above we say that a subtree is left-heavy if the balance factor is greater than zero. If the balance factor is less than zero then the subtree is right heavy. If the balance factor is zero then the tree is perfectly in balance. A tree is considered to be in balanced if the balance factor is -1, 0, or 1. Once the balance factor of a node in a tree is outside this range we will need to have a procedure to bring the tree back into balance.\n",
    "\n",
    "```{mermaid}\n",
    "flowchart TB;\n",
    "    a((4))\n",
    "    b((0))\n",
    "    c((5))\n",
    "    d((7))\n",
    "    e((2))\n",
    "    f((1))\n",
    "\n",
    "    a --> b\n",
    "    a --> c\n",
    "    c --> d\n",
    "    c --> e\n",
    "    e --> f\n",
    "```"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": []
  }
 ],
 "metadata": {
  "interpreter": {
   "hash": "dae0f32b50a4d158a1a751f5eb363916827e94eab0ac36188e044907592a3788"
  },
  "kernelspec": {
   "display_name": "Python 3.10.0 64-bit ('.venv': venv)",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.10.0"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 4
}
